################################################################################
# Copyright 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell cop-
# ies of the Software, and to permit persons to whom the Software is furnished
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IM-
# PLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNE-
# CTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
################################################################################

import itertools

from . import Properties
from . import Hardware
from . import Common
from . import Contractions
from .SolutionStructs import Solution as OriginalSolution
from .Utils import state

class SingleSolutionLibrary:
    Tag = 'Single'

    def __init__(self, solution):
        self.solution = solution

    @property
    def tag(self):
        return self.__class__.Tag

    def state(self):
        return {'type': self.tag, 'index': self.solution.index}

    def remapSolutionIndices(self,indexMap):
        pass

class GranularitySelectionLibrary:
    Tag = 'GranularitySelection'
    StateKeys = [('type', 'tag'), 'indices', 'exact']

    @classmethod
    def FromOriginalState(cls, d, indices):
        origTable = d[1]
        #indices = d[9]['TileSelectionIndices']
        #  entry = {'key': key, 'value': value, 'speed': row[1][1]}
        entries = []
        for row in origTable:
            try:
                index = row[1][0]
                key = list(row[0][0:4])
                entry = {'key': key, 'value': index }
                entries.append(entry)
            except KeyError:
                pass


        return cls(indices, entries)

    @property
    def tag(self):
        return self.__class__.Tag

    def merge(self, other):
        assert self.__class__ == other.__class__
        self.indices = list(set().union(self.indices, other.indices))
        self.exact = list(set().union(self.exact, other.exact))

    def __init__(self, indices, exact):
        self.indices = indices
        self.exact = exact

    def remapSolutionIndices(self,indexMap):
        for i in range(0, len(self.indices)):
            index = self.indices[i]
            if index in indexMap:
                self.indices[i] = indexMap[index]

class MatchingLibrary:
    Tag = 'Matching'
    StateKeys = [('type', 'tag'), 'properties', 'table', 'distance']

    @classmethod
    def FromOriginalState(cls, d, solutions, distance='Euclidean'):
        indices = d[0]
        origTable = d[1]

        propertyKeys = {
                2:lambda: Properties.Property('FreeSizeA', index=0),
                3:lambda: Properties.Property('FreeSizeB', index=0),
                #0:lambda: Properties.Property('BatchSize', index=0),
                1:lambda: Properties.Property('BoundSize', index=0)
            }

        properties = list([propertyKeys[i]() for i in indices if i in propertyKeys])
        keyOrder = [i for i,j in enumerate(indices) if j in propertyKeys]

        table = []

        for row in origTable:
            try:
                index = row[1][0]
                value = SingleSolutionLibrary(solutions[index])
                key = list([row[0][i] for i in keyOrder])
                #key = list(row[0][0:len(properties)])
                entry = {'key': key, 'value': value, 'speed': row[1][1]}
                table.append(entry)
            except KeyError:
                pass

        table.sort(key=lambda r: r['key'])

        return cls(properties, table, distance)

    @property
    def tag(self):
        return self.__class__.Tag

    def merge(self, other):
        assert self.__class__ == other.__class__ \
                and self.properties == other.properties \
                and self.distance == other.distance

        self.table += other.table

        self.table.sort(key=lambda r: r['key'])

    def remapSolutionIndices(self,indexMap):
        pass

    def __init__(self, properties, table, distance):
        self.properties = properties
        self.table = table
        self.distance = distance

class ProblemMapLibrary:
    Tag = 'ProblemMap'
    StateKeys = [('type', 'tag'), ('property', 'mappingProperty'), ('map', 'mapping')]

    def __init__(self, mappingProperty=None, mapping=None):
        self.mappingProperty = mappingProperty
        self.mapping = mapping

    @property
    def tag(self):
        return self.__class__.Tag

    def merge(self, other):
        assert self.__class__ == other.__class__ and self.tag == other.tag and self.mappingProperty == other.mappingProperty

        for key,value in list(other.mapping.items()):
            if key in self.mapping:
                self.mapping[key].merge(value)
            else:
                self.mapping[key] = value

    def remapSolutionIndices(self,indexMap):
        for key,value in list(self.mapping.items()):
            value.remapSolutionIndices(indexMap)

class PredicateLibrary:
    StateKeys = [('type', 'tag'), 'rows']

    def __init__(self, tag=None, rows=None):
        self.tag = tag
        if rows is None: rows = []
        self.rows = rows

    def merge(self, other):
        assert self.__class__ == other.__class__ and self.tag == other.tag

        rowdict = {r['predicate']: i  for i,r in enumerate(self.rows)}

        for row in other.rows:
            if row['predicate'] in rowdict:
                myRownum = rowdict[row['predicate']]
                self.rows[myRownum]['library'].merge(row['library'])
            else:
                self.rows.append(row)

        # Sort to ensure consistent fallback logic.
        self.rows.sort(key=lambda x: x['predicate'])

    def remapSolutionIndices(self,indexMap):
        for row in self.rows:
          row['library'].remapSolutionIndices(indexMap)


class MasterSolutionLibrary:
    StateKeys = ['solutions', 'library']

    @classmethod
    def FixSolutionIndices(cls, solutions):
        # fix missing and duplicate solution indices.
        try:
            maxSolutionIdx = max([s.index for s in solutions if s.index is not None])
        except ValueError:
            maxSolutionIdx = -1

        solutionsSoFar = set()
        for solution in solutions:
            if solution.index is None or solution.index in solutionsSoFar:
                maxSolutionIdx += 1
                solution.index = maxSolutionIdx
            else:
                solutionsSoFar.add(solution.index)
    @classmethod
    def FromOriginalState(cls, d, origSolutions, solutionClass=Contractions.Solution, libraryOrder = None):
        if libraryOrder is None:
            libraryOrder = ['Hardware', 'OperationIdentifier', 'PerformanceMetric', 'Fp16AltImpl', 'Predicates', 'Matching']

        deviceSection = d[1:4]
        origProblemType = d[4]
        #origSolutions = d[5]
        origLibrary = d[6:8]

        if len(d) > 9 and d[9]:
            # It's a Granularity library
            assert libraryOrder[-1] == 'Matching'
            libraryOrder[-1] = 'Granularity'

        perfMetric = 'DeviceEfficiency'
        if len(d) > 10 and d[10]:
            perfMetric = d[10]

        fp16AltImpl = False
        if len(d) > 11 and d[11]:
            fp16AltImpl = True

        matching = 'Euclidean'
        if len(d) > 12 and d[12]:
            matching = d[12]

        problemType = Contractions.ProblemType.FromOriginalState(origProblemType)

        allSolutions = [solutionClass.FromSolutionStruct(s, deviceSection) for s in origSolutions]
        cls.FixSolutionIndices(allSolutions)

        solutions = {s.index: s for s in allSolutions}

        for libName in reversed(libraryOrder):
            if libName == 'Matching':
                if matching == 'Equality':
                    predicate = Properties.Predicate(tag='EqualityMatching')
                else:
                    predicate = Properties.Predicate(tag='TruePred')

                matchingLib = MatchingLibrary.FromOriginalState( \
                        origLibrary, allSolutions, matching)
                library = PredicateLibrary(tag='Problem')
                library.rows.append({'predicate': predicate, 'library': matchingLib})

            elif libName == 'Granularity':
                selectionIndices = d[9]['TileSelectionIndices']
                library = GranularitySelectionLibrary.FromOriginalState(origLibrary, selectionIndices)

            elif libName == 'Hardware':

                if isinstance(deviceSection[1], dict):
                    architectureProps = deviceSection[1]
                    assert 'Architecture' in architectureProps, 'Invalid device section [1]'
                    assert 'CUCount' in architectureProps, 'Invalid device section [1]'
                    devicePart = architectureProps['Architecture']
                    cuCount = architectureProps['CUCount']
                else:
                    devicePart = deviceSection[1]
                    cuCount = None

                newLib = PredicateLibrary(tag='Hardware')
                if devicePart == 'fallback':
                    pred = Hardware.HardwarePredicate('TruePred')
                else:
                    pred = Hardware.HardwarePredicate.FromHardware(Common.gfxArch(devicePart), cuCount)

                newLib.rows.append({'predicate': pred, 'library': library})
                library = newLib

            elif libName == 'Predicates':
                predicates = problemType.predicates(includeBatch=True, includeType=True)
                predicate = Contractions.ProblemPredicate.And(predicates)

                newLib = PredicateLibrary(tag='Problem')
                newLib.rows.append({'predicate': predicate, 'library': library})
                library = newLib

            elif libName == 'OperationIdentifier':
                operationID = problemType.operationIdentifier
                prop = Properties.Property('OperationIdentifier')
                mapping = {operationID: library}

                newLib = ProblemMapLibrary(prop, mapping)
                library = newLib

            elif libName == 'PerformanceMetric':
                if perfMetric != 'DeviceEfficiency':
                    predicate = Properties.Predicate(tag=perfMetric)
                else:
                    predicate = Properties.Predicate(tag='TruePred')
                newLib = PredicateLibrary(tag='Problem')
                newLib.rows.append({'predicate': predicate, 'library': library})
                library = newLib

            elif libName == 'Fp16AltImpl':
                if fp16AltImpl:
                    predicate = Properties.Predicate(tag='Fp16AltImpl')
                else:
                    predicate = Properties.Predicate(tag='TruePred')
                newLib = PredicateLibrary(tag='Problem')
                newLib.rows.append({'predicate': predicate, 'library': library})
                library = newLib

            else:
                raise ValueError('Unknown value ' + libName)

        rv = cls(solutions, library)
        return rv

    @classmethod
    def BenchmarkingLibrary(cls, solutions):
        solutionObjs = list([Contractions.Solution.FromOriginalState(s._state) for s in solutions])
        cls.FixSolutionIndices(solutionObjs)

        predRows = list([{'predicate': s.problemPredicate, 'library': SingleSolutionLibrary(s)} for s in solutionObjs])
        library = PredicateLibrary(tag='Problem', rows=predRows)

        solutionMap = {s.index: s for s in solutionObjs}

        return cls(solutionMap, library)

    def __init__(self, solutions, library, version=None):
        self.solutions = solutions
        self.library = library
        self.version = version

    def state(self):
        rv = {'solutions': state(iter(list(self.solutions.values()))),
              'library': state(self.library)}

        if self.version is not None:
            rv['version'] = self.version
        return rv

    def applyNaming(self, naming=None):
        if naming is None:
            #allSolutions = itertools.chain(iter(list(self.solutions.values())), iter(list(self.sourceSolutions.values())))
            kernels = list(itertools.chain(*[s.originalSolution.getKernels() for s in self.solutions.values()]))
            naming = OriginalSolution.getMinNaming(kernels)

        for s in list(self.solutions.values()):
            s.name = OriginalSolution.getNameMin(s.originalSolution.getKernels()[0], naming)

    def merge(self, other):
        assert self.__class__ == other.__class__

        curIndex = max(self.solutions.keys()) + 1

        reIndexMap = {}
        for k,s in other.solutions.items():
            reIndexMap[s.index] = curIndex
            s.index = curIndex
            self.solutions[curIndex] = s
            curIndex += 1

        other.library.remapSolutionIndices(reIndexMap)

        self.library.merge(other.library)

    @property
    def cpp_base_class(self):
        return 'SolutionLibrary<ContractionProblem, ContractionSolution>'

    @property
    def cpp_class(self):
        return 'MasterSolutionLibrary<ContractionProblem, ContractionSolution>'
